import threading
import time

import requests
from requests import exceptions as requests_exceptions

from hackle import version
from . import logger as hackle_logger
from . import workspace
from .commons import enums


class WorkspaceFetcher(object):
    WORKSPACE_URL = enums.Default.WORKSPACE_URL

    def __init__(self, sdk_key=None, update_interval=None, blocking_timeout=None, logger=None):
        self.sdk_key = sdk_key
        self.update_interval = update_interval or enums.Default.DEFAULT_POLL_INTERVAL
        self.blocking_timeout = blocking_timeout or enums.Default.DEFAULT_BLOCKING_TIMEOUT
        self.logger = hackle_logger.adapt_logger(logger or hackle_logger.NoOpLogger())
        self._workspace_fetch_ready_event = threading.Event()
        self._polling_thread = threading.Thread(target=self._run)
        self._polling_thread.setDaemon(True)
        self._polling_thread.start()

    def _handle_response(self, response):
        try:
            response.raise_for_status()
        except requests_exceptions.RequestException as err:
            self.logger.error('Failed to pool Workspace: {}'.format(err))
            return

        self._set_config(response.content)

    def fetch(self):
        request_headers = {enums.Default.SDK_KEY_HEADER: self.sdk_key,
                           enums.Default.SDK_NAME_HEADER: 'python-sdk',
                           enums.Default.SDK_VERSION_HEADER: version.__version__}
        try:
            response = requests.get(
                self.WORKSPACE_URL,
                headers=request_headers,
                timeout=enums.Default.REQUEST_TIMEOUT
            )
        except requests_exceptions.RequestException as err:
            self.logger.error('Failed to pool Workspace: {}'.format(err))
            return

        self._handle_response(response)

    @property
    def is_running(self):
        return self._polling_thread.is_alive()

    def _run(self):
        try:
            while self.is_running:
                self.fetch()
                time.sleep(self.update_interval)
        except (OSError, OverflowError) as err:
            self.logger.error('Error : {}'.format(str(err)))
            raise

    def start(self):
        if not self.is_running:
            self._polling_thread.start()

    def _set_config(self, content):
        if content or self._workspace_fetch_ready_event.is_set():
            self._workspace = workspace.Workspace(content, self.logger)
            self._workspace_fetch_ready_event.set()

    def get_config(self):
        self._workspace_fetch_ready_event.wait(self.blocking_timeout)
        if hasattr(self, '_workspace'):
            return self._workspace
        return None

    def stop(self):
        if self.is_running:
            self._polling_thread.join(self.blocking_timeout)



from . import decider
from . import exceptions as hackle_exceptions
from . import logger as _logging
from .commons import validator, enums
from .event import event_processor, event as event_entity
from .event.event_dispatcher import EventDispatcher
from .workspace_fetcher import WorkspaceFetcher


def __singleton(cls):
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper


@__singleton
class Client(object):
    def __init__(
            self,
            sdk_key=None,
            logger=None
    ):
        if sdk_key is None:
            raise hackle_exceptions.RequiredParameterException('sdk_key must not be empty')

        self.event_dispatcher = EventDispatcher()
        self.logger = _logging.adapt_logger(logger or _logging.NoOpLogger())
        self.event_processor = event_processor.BatchEventProcessor(sdk_key, self.event_dispatcher, self.logger)
        self.decider = decider.Decider(self.logger)
        self.workspace_fetcher = WorkspaceFetcher(sdk_key, logger=self.logger)

    def close(self):
        self.workspace_fetcher.stop()
        self.event_processor.stop()

    def __exit__(self):
        self.close()

    def variation(self, experiment_key, user, default_variation='A'):
        """
        Decide the variation to expose to the user for experiment.

        This method return the "A" if:
            - The experiment key is invalid
            - The experiment has not started yet
            - The user is not allocated to the experiment
            - The decided variation has been dropped

        :param int experiment_key: the unique key of the experiment.
        :param hackle.model.User user: the identifier of your customer (e.g. user_email, account_id, etc.) MUST NOT be
         null.
        :param str default_variation: the default variation of the experiment. MUST NOT be null.
        :return: the decided variation for the user, or the default variation.
        """
        if not validator.is_non_zero_and_empty_int(experiment_key):
            self.logger.error('Experiment Key Is Not Empty : {}'.format(experiment_key))
            return default_variation

        if not validator.is_valid_user(user):
            self.logger.error('User is not valid. user\'s type must be hackle.model.User and user.id\'s type must be '
                              'string_types : {}'.format(user))
            return default_variation

        config = self.workspace_fetcher.get_config()
        if not config:
            self.logger.error('Invalid Workspace. Hackle instance is not valid. {}'.format('variation'))
            return default_variation

        experiment = config.get_experiment_from_key(experiment_key)

        if not experiment:
            self.logger.info('Experiment Key "%s" is invalid. Not Allocated.' % experiment_key)
            return default_variation

        decision = self.decider.decide(experiment, user)
        if not decision:
            return default_variation

        if decision.type == enums.DecisionType.FORCED:
            return decision.variation
        elif decision.type == enums.DecisionType.NATURAL:
            self.event_processor.process(
                event_entity.ExposureEvent(user, decision.experiment.id, decision.experiment.key, decision.variation.id,
                                           decision.variation.key))
            return decision.variation.key
        return default_variation

    def track(self, event, user):
        """
        Records the event performed by the user with additional numeric value.

        :param hackle.model.Event event: the unique key of the event. MUST NOT be null.
        :param user: the identifier of user that performed the event. MUST NOT be null.
        :return:
        """
        if not validator.is_valid_event(event):
            self.logger.error('Event is not valid. Event must be hackle.model.event and event.id\'s type must be '
                              'string_types. value\'s type must be numeric. '
                              ': {}'.format(event))
            return

        if not validator.is_valid_user(user):
            self.logger.error('User is not valid. user\'s type must be hackle.model.User and user.id\'s type must be '
                              'string_types : {}'.format(user))
            return

        config = self.workspace_fetcher.get_config()
        if not config:
            self.logger.error('Invalid Workspace. Hackle instance is not valid. {}'.format('event'))
            return

        event_type = config.get_event(event.key)
        self.event_processor.process(
            event_entity.TrackEvent(user, event_type, event)
        )

        return



import json

from . import entities


class Workspace(object):
    def __init__(self, data, logger):
        self.variation_id_map = None
        json_data = json.loads(data)
        self.logger = logger
        self.logger.debug('data : {}'.format(json_data))

        self.workspace = json_data.get('workspace')
        self.experiments = json_data.get('experiments', [])
        self.completedExperiments = json_data.get('completedExperiments', [])
        self.buckets = json_data.get('buckets', [])
        self.events = json_data.get('events', [])

        self.experiment_key_map = {}
        self.completed_experiment_key_map = {}
        self.bucket_id_map = {}
        self.event_key_map = {}

        for bucket in self.buckets:
            slots = []
            for slot in bucket['slots']:
                slots.append(entities.Slot(slot['startInclusive'], slot['endExclusive'], slot['variationId']))
            self.bucket_id_map[bucket['id']] = entities.Bucket(bucket['seed'], bucket['slotSize'], slots)

        for event in self.events:
            self.event_key_map[str(event['key'])] = entities.Event(event['id'], event['key'])

        for experiment in self.experiments:
            variations = {}
            for variation in experiment['variations']:
                variations[variation['id']] = entities.Variation(variation['id'], variation['key'],
                                                                 variation['status'] == 'DROPPED')

                self.experiment_key_map[experiment['key']] = entities.RunningExperiment(experiment['id'],
                                                                                        experiment['key'],
                                                                                        self.bucket_id_map[
                                                                                            experiment[
                                                                                                'bucketId']],
                                                                                        variations,
                                                                                        experiment['execution']
                                                                                        )

        for completedExperiment in self.completedExperiments:
            self.completed_experiment_key_map[completedExperiment['experimentKey']] = entities.CompletedExperiment(
                completedExperiment['experimentId'],
                completedExperiment['experimentKey'],
                completedExperiment['winnerVariationKey']
            )
            self.experiment_key_map[completedExperiment['experimentKey']] = entities.CompletedExperiment(
                completedExperiment['experimentId'],
                completedExperiment['experimentKey'],
                completedExperiment['winnerVariationKey']
            )

    @staticmethod
    def _generate_key_map(entity_list, key, entity_class):
        key_map = {}
        for obj in entity_list:
            print(type(obj))
            key_map[obj[key]] = entity_class(**obj)

        return key_map

    def get_experiment_from_key(self, experiment_key):
        experiment = self.experiment_key_map.get(experiment_key)
        if experiment:
            return experiment

        self.logger.error('Experiment Key %s is not defined.' % experiment_key)

        return None

    def get_completed_experiment_from_key(self, experiment_key):
        completed_experiment = self.completed_experiment_key_map.get(experiment_key)

        if completed_experiment:
            return completed_experiment

        return None

    def get_event(self, event_key):
        event = self.event_key_map.get(event_key)

        if event:
            return event
        else:
            return entities.Event(0, event_key)



class User(object):
    def __init__(self, id, properties):
        self.id = id
        self.properties = properties


class Event(object):
    def __init__(self, key, value, properties):
        self.key = key
        self.value = value
        self.properties = properties


class Hackle:
    @staticmethod
    def user(id, **kwargs):
        return User(id, kwargs)

    @staticmethod
    def event(key, value=None, **kwargs):
        return Event(key, value, kwargs)



import abc

ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})


class BaseExperiment(ABC):
    def __init__(self, id, key):
        self.id = id
        self.key = key


class RunningExperiment(BaseExperiment):
    def __init__(self, id, key, bucket, variations, execution, **kwargs):
        super(RunningExperiment, self).__init__(id, key)
        self.bucket = bucket
        self.variations = variations
        user_overrides = {}
        for userOverride in execution['userOverrides']:
            user_overrides[userOverride['userId']] = self.variations.get(userOverride['variationId'])
        self.userOverrides = user_overrides

    def __repr__(self):
        return '(id={}, key={}, bucket={}, variations={}, userOverrides={})'.format(self.id, self.key,
                                                                                    self.bucket, self.variations,
                                                                                    self.userOverrides)


class CompletedExperiment(BaseExperiment):
    def __init__(self, id, key, winner_variation_key):
        super(CompletedExperiment, self).__init__(id, key)
        self.winnerVariationKey = winner_variation_key


class Variation(object):
    def __init__(self, id, key, is_dropped, **kwargs):
        self.id = id
        self.key = key
        self.isDropped = is_dropped


class UserOverride(object):
    def __init__(self, user_id, variation, **kwargs):
        self.userId = user_id,
        self.variation = variation


class Bucket(object):
    def __init__(self, seed, slot_size, slots, **kwargs):
        self.seed = seed
        self.slotSize = slot_size
        self.slots = slots


class Slot(object):
    def __init__(self, start_inclusive, end_exclusive, variation_id, **kwargs):
        self.startInclusive = start_inclusive
        self.endExclusive = end_exclusive
        self.variationId = variation_id

    def contains(self, slot_number):
        return (self.startInclusive <= slot_number) and (slot_number < self.endExclusive)


class Event(object):
    def __init__(self, id, key):
        self.id = id
        self.key = key
